﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.VehicleRouting.Encodings;
using HeuristicLab.Problems.VehicleRouting.Encodings.Potvin;
using HeuristicLab.Parameters;
using HeuristicLab.Optimization;

namespace HeuristicLab.Problems.VehicleRouting {
  /// <summary>
  /// An operator that relinks paths between vehicle routing solutions.
  /// </summary>
  [Item("VRPPathRelinker", "An operator that relinks paths between vehicle routing solutions.")]
  [StorableClass]
  public sealed class VRPPathRelinker : PathRelinker, IStochasticOperator {
    #region Parameters
    public ILookupParameter<DoubleMatrix> CoordinatesParameter {
      get {
        if (Parameters.ContainsKey("Coordinates"))
          return (ILookupParameter<DoubleMatrix>)Parameters["Coordinates"];
        else
          return null;
      }
    }
    public ILookupParameter<DoubleMatrix> DistanceMatrixParameter {
      get {
        if (Parameters.ContainsKey("DistanceMatrix"))
          return (ILookupParameter<DoubleMatrix>)Parameters["DistanceMatrix"];
        else
          return null;
      }
    }
    public ILookupParameter<BoolValue> UseDistanceMatrixParameter {
      get {
        if (Parameters.ContainsKey("UseDistanceMatrix"))
          return (ILookupParameter<BoolValue>)Parameters["UseDistanceMatrix"];
        else
          return null;
      }
    }
    public ILookupParameter<IntValue> VehiclesParameter {
      get {
        if (Parameters.ContainsKey("Vehicles"))
          return (ILookupParameter<IntValue>)Parameters["Vehicles"];
        else
          return null;
      }
    }
    public ILookupParameter<DoubleValue> CapacityParameter {
      get {
        if (Parameters.ContainsKey("Capacity"))
          return (ILookupParameter<DoubleValue>)Parameters["Capacity"];
        else
          return null;
      }
    }
    public ILookupParameter<DoubleArray> DemandParameter {
      get {
        if (Parameters.ContainsKey("Demand"))
          return (ILookupParameter<DoubleArray>)Parameters["Demand"];
        else
          return null;
      }
    }
    public ILookupParameter<DoubleArray> ReadyTimeParameter {
      get {
        if (Parameters.ContainsKey("ReadyTime"))
          return (ILookupParameter<DoubleArray>)Parameters["ReadyTime"];
        else
          return null;
      }
    }
    public ILookupParameter<DoubleArray> DueTimeParameter {
      get {
        if (Parameters.ContainsKey("DueTime"))
          return (ILookupParameter<DoubleArray>)Parameters["DueTime"];
        else
          return null;
      }
    }
    public ILookupParameter<DoubleArray> ServiceTimeParameter {
      get {
        if (Parameters.ContainsKey("ServiceTime"))
          return (ILookupParameter<DoubleArray>)Parameters["ServiceTime"];
        else
          return null;
      }
    }
    public ILookupParameter<DoubleValue> FleetUsageFactor {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalFleetUsageFactor"]; }
    }
    public ILookupParameter<DoubleValue> TimeFactor {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalTimeFactor"]; }
    }
    public ILookupParameter<DoubleValue> DistanceFactor {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalDistanceFactor"]; }
    }
    public ILookupParameter<DoubleValue> OverloadPenalty {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalOverloadPenalty"]; }
    }
    public ILookupParameter<DoubleValue> TardinessPenalty {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalTardinessPenalty"]; }
    }
    public ILookupParameter<IRandom> RandomParameter {
      get { return (LookupParameter<IRandom>)Parameters["Random"]; }
    }

    public IValueParameter<IntValue> SampleSizeParameter {
      get { return (IValueParameter<IntValue>)Parameters["SampleSize"]; }
    }
    public IValueParameter<IntValue> IterationsParameter {
      get { return (IValueParameter<IntValue>)Parameters["Iterations"]; }
    }
    #endregion

    [StorableConstructor]
    private VRPPathRelinker(bool deserializing) : base(deserializing) { }
    private VRPPathRelinker(VRPPathRelinker original, Cloner cloner) : base(original, cloner) { }
    public VRPPathRelinker() : base() {
      Parameters.Add(new LookupParameter<DoubleMatrix>("Coordinates", "The coordinates of the cities."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("DistanceMatrix", "The matrix which contains the distances between the cities."));
      Parameters.Add(new LookupParameter<BoolValue>("UseDistanceMatrix", "True if a distance matrix should be calculated and used for evaluation, otherwise false."));
      Parameters.Add(new LookupParameter<IntValue>("Vehicles", "The number of vehicles."));
      Parameters.Add(new LookupParameter<DoubleValue>("Capacity", "The capacity of each vehicle."));
      Parameters.Add(new LookupParameter<DoubleArray>("Demand", "The demand of each customer."));
      Parameters.Add(new LookupParameter<DoubleArray>("ReadyTime", "The ready time of each customer."));
      Parameters.Add(new LookupParameter<DoubleArray>("DueTime", "The due time of each customer."));
      Parameters.Add(new LookupParameter<DoubleArray>("ServiceTime", "The service time of each customer."));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalFleetUsageFactor", "The fleet usage factor considered in the evaluation."));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalTimeFactor", "The time factor considered in the evaluation."));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalDistanceFactor", "The distance factor considered in the evaluation."));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalOverloadPenalty", "The overload penalty considered in the evaluation."));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalTardinessPenalty", "The tardiness penalty considered in the evaluation."));
      Parameters.Add(new LookupParameter<IRandom>("Random", "The pseudo random number generator which should be used for stochastic manipulation operators."));
      Parameters.Add(new ValueParameter<IntValue>("SampleSize", new IntValue(10)));
      Parameters.Add(new ValueParameter<IntValue>("Iterations", new IntValue(50)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new VRPPathRelinker(this, cloner);
    }

    private static int MatchingCities(Tour tour1, Tour tour2) {
      return tour1.Cities.Intersect(tour2.Cities).Count();
    }

    private static PotvinEncoding MatchTours(PotvinEncoding initiator, PotvinEncoding guide) {
      PotvinEncoding result = new PotvinEncoding();

      List<bool> used = new List<bool>();
      for (int i = 0; i < initiator.Tours.Count; i++) {
        used.Add(false);
      }

      for (int i = 0; i < guide.Tours.Count; i++) {
        int bestMatch = -1;
        int bestTour = -1;

        for (int j = 0; j < initiator.Tours.Count; j++) {
          if (!used[j]) {
            int match = MatchingCities(guide.Tours[i], initiator.Tours[j]);
            if (match > bestMatch) {
              bestMatch = match;
              bestTour = j;
            }
          }
        }

        if (bestTour != -1) {
          result.Tours.Add(initiator.Tours[bestTour]);
          used[bestTour] = true;
        }
      }

      for (int i = 0; i < initiator.Tours.Count; i++) {
        if (!used[i])
          result.Tours.Add(initiator.Tours[i]);
      }

      return result;
    }

    #region moves
    public static PotvinEncoding RouteBasedXOver(PotvinEncoding initiator, PotvinEncoding guide, IRandom random,
      IntValue vehicles, DoubleArray dueTimeArray,
      DoubleArray serviceTimeArray, DoubleArray readyTimeArray, DoubleArray demandArray, DoubleValue capacity,
      DoubleMatrix coordinates, IParameter distanceMatrix, BoolValue useDistanceMatrix) {
        return PotvinRouteBasedCrossover.Apply(random, initiator, guide, false,
          dueTimeArray, serviceTimeArray, readyTimeArray, demandArray,
          capacity, coordinates, distanceMatrix, useDistanceMatrix);
    }

    public static PotvinEncoding SequenceBasedXOver(PotvinEncoding initiator, PotvinEncoding guide, IRandom random,
     IntValue vehicles, DoubleArray dueTimeArray,
     DoubleArray serviceTimeArray, DoubleArray readyTimeArray, DoubleArray demandArray, DoubleValue capacity,
     DoubleMatrix coordinates, IParameter distanceMatrix, BoolValue useDistanceMatrix) {
      return PotvinSequenceBasedCrossover.Apply(random, initiator, guide, false,
        dueTimeArray, serviceTimeArray, readyTimeArray, demandArray,
        capacity, coordinates, distanceMatrix, useDistanceMatrix);
    }


    public static void GuidedRelocateMove(PotvinEncoding initiator, PotvinEncoding guide, IRandom random) {
      List<int> cities = new List<int>();
      foreach (Tour tour in initiator.Tours) {
        foreach (int city in tour.Cities) {
          Tour guideTour = guide.Tours.First(t => t.Cities.Contains(city));
          if (guide.Tours.IndexOf(guideTour) != initiator.Tours.IndexOf(tour)) {
            cities.Add(city);
          }
        }
      }

      if (cities.Count == 0) {
        RelocateMove(initiator, random);
      } else {
        int city = cities[random.Next(cities.Count)];
        Tour tour = initiator.Tours.First(t => t.Cities.Contains(city));
        tour.Cities.Remove(city);

        Tour guideTour = guide.Tours.First(t => t.Cities.Contains(city));
        int guideTourIndex = guide.Tours.IndexOf(guideTour);

        if (guideTourIndex < initiator.Tours.Count) {
          Tour tour2 = initiator.Tours[guideTourIndex];

          int guideIndex = guideTour.Cities.IndexOf(city);
          if (guideIndex == 0) {
            tour2.Cities.Insert(0, city);
          } else {
            int predecessor = guideTour.Cities[guideIndex - 1];
            int initIndex = tour2.Cities.IndexOf(predecessor);
            if (initIndex != -1) {
              tour2.Cities.Insert(initIndex + 1, city);
            } else {
              if (guideIndex == guideTour.Cities.Count - 1) {
                tour2.Cities.Insert(tour2.Cities.Count, city);
              } else {
                int sucessor = guideTour.Cities[guideIndex + 1];
                initIndex = tour2.Cities.IndexOf(sucessor);
                if (initIndex != -1) {
                  tour2.Cities.Insert(initIndex, city);
                } else {
                  tour2.Cities.Insert(random.Next(tour2.Cities.Count + 1), city);
                }
              }
            }
          }
        } else {
          Tour tour2 = new Tour();
          tour2.Cities.Add(city);
          initiator.Tours.Add(tour2);
        }

        if (tour.Cities.Count == 0)
          initiator.Tours.Remove(tour);
      }
    }

    public static void RelocateMove(PotvinEncoding individual, IRandom random) {
      int cities = individual.Cities;
      int city = 1 + random.Next(cities);
      Tour originalTour = individual.Tours.Find(t => t.Cities.Contains(city));
      //consider creating new route
      individual.Tours.Add(new Tour());

      int position = 1 + random.Next(cities + individual.Tours.Count - 1);
      if (position >= city) {
        position++;
      }

      int originalPosition = originalTour.Cities.IndexOf(city);
      originalTour.Cities.RemoveAt(originalPosition);

      Tour insertionTour;
      int insertionPosition;
      if (position <= cities) {
        insertionTour = individual.Tours.Find(t => t.Cities.Contains(position));
        insertionPosition = insertionTour.Cities.IndexOf(position) + 1;
      } else {
        insertionTour = individual.Tours[position - cities - 1];
        insertionPosition = 0;
      }

      insertionTour.Cities.Insert(insertionPosition, city);

      individual.Tours.RemoveAll(t => t.Cities.Count == 0);
    }

    public static void ExchangeMove(PotvinEncoding individual, IRandom random) {
      if (individual.Tours.Count > 1) {
        int tour1Idx = random.Next(individual.Tours.Count);
        int tour2Idx = random.Next(individual.Tours.Count - 1);
        if (tour2Idx >= tour1Idx)
          tour2Idx++;

        Tour tour1 = individual.Tours[tour1Idx];
        Tour tour2 = individual.Tours[tour2Idx];

        int index1 = random.Next(tour1.Cities.Count);
        int city1 = tour1.Cities[index1];

        int index2 = random.Next(tour2.Cities.Count);
        int city2 = tour2.Cities[index2];

        tour1.Cities.RemoveAt(index1);
        tour1.Cities.Insert(index1, city2);

        tour2.Cities.RemoveAt(index2);
        tour2.Cities.Insert(index2, city1);
      }
    }

    public static void TwoOptMove(PotvinEncoding individual, IRandom random) {
      List<Tour> tours = individual.Tours.FindAll(t => t.Cities.Count >= 4);

      if (tours.Count > 0) {
        int tourIdx = random.Next(tours.Count);
        Tour tour = tours[tourIdx];

        int a;
        if (tour.Cities.Count == 4) {
          a = 0;
        } else if (tour.Cities.Count == 5) {
          int idx = random.Next(4);
          if (idx >= 2)
            idx++;
          a = idx;
        } else {
          a = random.Next(tour.Cities.Count);
        }

        int b;
        List<int> indices = new List<int>();
        for (int i = 0; i < tour.Cities.Count; i++) {
          if (Math.Abs(i - a) > 2) {
            indices.Add(i);
          }
        }
        b = indices[random.Next(indices.Count)];

        if (b < a) {
          int tmp = a;
          a = b;
          b = tmp;
        }

        int index = a + 1;
        int count = b - a - 1;
        List<int> segment = tour.Cities.GetRange(index, count);
        tour.Cities.RemoveRange(index, count);
        segment.Reverse();
        tour.Cities.InsertRange(index, segment);
      }
    }

    public static void TwoOptStarMove(PotvinEncoding individual, IRandom random) {
      //consider creating new tour
      individual.Tours.Add(new Tour());

      int route1Idx = random.Next(individual.Tours.Count);
      int route2Idx = random.Next(individual.Tours.Count - 1);
      if (route2Idx >= route1Idx)
        route2Idx++;

      Tour route1 = individual.Tours[route1Idx];
      Tour route2 = individual.Tours[route2Idx];

      int x1 = random.Next(route1.Cities.Count + 1);
      int x2 = random.Next(route2.Cities.Count + 1);

      int count = route1.Cities.Count - x1;
      List<int> segmentX1 = new List<int>();
      if (count > 0) {
        segmentX1 = route1.Cities.GetRange(x1, count);
        route1.Cities.RemoveRange(x1, count);
      }

      count = route2.Cities.Count - x2;
      List<int> segmentX2 = new List<int>();
      if (count > 0) {
        segmentX2 = route2.Cities.GetRange(x2, count);
        route2.Cities.RemoveRange(x2, count);
      }

      route1.Cities.AddRange(segmentX2);
      route2.Cities.AddRange(segmentX1);

      individual.Tours.RemoveAll(t => t.Cities.Count == 0);
    }

    public static void OrOptMove(PotvinEncoding individual, IRandom random) {
      List<Tour> tours = individual.Tours.FindAll(t => t.Cities.Count >= 2);

      if (tours.Count > 0) {
        int tourIdx = random.Next(tours.Count);
        Tour tour = tours[tourIdx];

        int segmentStart = random.Next(tour.Cities.Count);
        int segmentLength;
        if (segmentStart == 0) {
          segmentLength = 1 + random.Next(tour.Cities.Count - 1);
        } else {
          segmentLength = 1 + random.Next(tour.Cities.Count - segmentStart);
        }

        List<int> segment = tour.Cities.GetRange(segmentStart, segmentLength);
        tour.Cities.RemoveRange(segmentStart, segmentLength);
        int newPos;
        if (tour.Cities.Count == 1)
          newPos = 0;
        else
          newPos = random.Next(tour.Cities.Count - 1);

        if (newPos >= segmentStart)
          newPos++;
        tour.Cities.InsertRange(newPos, segment);
      }
    }
    #endregion

    private static IList<IItem> ChooseSelection(IList<PotvinEncoding> solutions, PercentValue n) {
      IList<IItem> selection = new List<IItem>();
      if (solutions.Count > 0) {
        int noSol = (int)(solutions.Count * n.Value);
        if (noSol <= 0) noSol++;
        double stepSize = (double)solutions.Count / (double)noSol;
        for (int i = 0; i < noSol; i++)
          selection.Add(solutions.ElementAt((int)((i + 1) * stepSize - stepSize * 0.5)));
      }

      return selection;
    }

    public static ItemArray<IItem> Apply(PotvinEncoding initiator, PotvinEncoding guide, PercentValue n, int sampleSize, int iterations, IRandom rand,
      IntValue vehicles, DoubleArray dueTimeArray,
      DoubleArray serviceTimeArray, DoubleArray readyTimeArray, DoubleArray demandArray, DoubleValue capacity,
      DoubleValue fleetUsageFactor, DoubleValue timeFactor, DoubleValue distanceFactor, DoubleValue overloadPenalty, DoubleValue tardinessPenalty,
      DoubleMatrix coordinates, IParameter distanceMatrix, BoolValue useDistanceMatrix) {
     
      double sigma = 1.5;

      DoubleValue currentOverloadPenalty = new DoubleValue(overloadPenalty.Value);
      DoubleValue currentTardinessPenalty = new DoubleValue(tardinessPenalty.Value);

      PotvinEncoding current = MatchTours(initiator, guide);
      double currentSimilarity = VRPSimilarityCalculator.CalculateSimilarity(current, guide);

      IList<PotvinEncoding> solutions = new List<PotvinEncoding>();
      int i = 0;
      while(i < iterations && currentSimilarity != 1.0) {
        TourEvaluation currentEval = VRPEvaluator.Evaluate(current,
          vehicles, dueTimeArray, serviceTimeArray, readyTimeArray,
          demandArray, capacity, fleetUsageFactor, timeFactor, distanceFactor,
          currentOverloadPenalty, currentTardinessPenalty, 
          coordinates, distanceMatrix, useDistanceMatrix);
        currentSimilarity = VRPSimilarityCalculator.CalculateSimilarity(current, guide);

        if (currentSimilarity < 1.0) {
          for (int sample = 0; sample < sampleSize; sample++) {
            PotvinEncoding next = current.Clone() as PotvinEncoding;

            int neighborhood = rand.Next(4);
            switch (neighborhood) {
              case 0: next = RouteBasedXOver(next, guide, rand, 
                vehicles, dueTimeArray, serviceTimeArray, readyTimeArray, demandArray, 
                capacity, coordinates, distanceMatrix, useDistanceMatrix);
                break;
              case 1: next = SequenceBasedXOver(next, guide, rand,
                vehicles, dueTimeArray, serviceTimeArray, readyTimeArray, demandArray,
                capacity, coordinates, distanceMatrix, useDistanceMatrix);
                  break;
              case 2: TwoOptMove(next, rand);
                break;
              case 3: GuidedRelocateMove(next, guide, rand);
                break;
            }

            next = MatchTours(next, guide);

            TourEvaluation nextEval = VRPEvaluator.Evaluate(next,
              vehicles, dueTimeArray, serviceTimeArray, readyTimeArray,
              demandArray, capacity, fleetUsageFactor, timeFactor, distanceFactor,
              currentOverloadPenalty, currentTardinessPenalty,
              coordinates, distanceMatrix, useDistanceMatrix);
            double nextSimilarity = VRPSimilarityCalculator.CalculateSimilarity(next, guide);

            if (nextSimilarity > currentSimilarity && nextEval.Quality <= currentEval.Quality) {
              current = next;
              solutions.Add(current);
              break;
            }
          }

          if (currentEval.Overload > 0)
            currentOverloadPenalty.Value *= sigma;
          else
            currentOverloadPenalty.Value /= sigma;

          if (currentEval.Tardiness > 0)
            currentTardinessPenalty.Value *= sigma;
          else
            currentTardinessPenalty.Value /= sigma;

          i++;
        }
      }

      return new ItemArray<IItem>(ChooseSelection(solutions, n));
    }

    protected override ItemArray<IItem> Relink(ItemArray<IItem> parents, PercentValue n) {
      if (parents.Length != 2)
        throw new ArgumentException("The number of parents is not equal to 2.");

      if(!(parents[0] is PotvinEncoding))
        parents[0] = PotvinEncoding.ConvertFrom(parents[0] as IVRPEncoding, DistanceMatrixParameter);
      if (!(parents[1] is PotvinEncoding))
        parents[1] = PotvinEncoding.ConvertFrom(parents[1] as IVRPEncoding, DistanceMatrixParameter);

      return Apply(parents[0] as PotvinEncoding, parents[1] as PotvinEncoding, n, 
        SampleSizeParameter.Value.Value, IterationsParameter.Value.Value, RandomParameter.ActualValue,
        VehiclesParameter.ActualValue, DueTimeParameter.ActualValue, ServiceTimeParameter.ActualValue, ReadyTimeParameter.ActualValue,
        DemandParameter.ActualValue, CapacityParameter.ActualValue,
        FleetUsageFactor.ActualValue, TimeFactor.ActualValue, DistanceFactor.ActualValue, OverloadPenalty.ActualValue, TardinessPenalty.ActualValue,
        CoordinatesParameter.ActualValue, DistanceMatrixParameter, UseDistanceMatrixParameter.ActualValue);
    }
  }
}
