﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.VehicleRouting.Encodings;
using HeuristicLab.Problems.VehicleRouting.Encodings.Potvin;
using HeuristicLab.Data;
using HeuristicLab.Parameters;

namespace HeuristicLab.Problems.VehicleRouting {
  /// <summary>
  /// An operator that performs similarity calculation between two vehicle routing solutions.
  /// </summary>
  [Item("VRPSimilarityCalculator", "An operator that performs similarity calculation between two vehicle routing solutions.")]
  [StorableClass]
  public sealed class VRPSimilarityCalculator : SimilarityCalculator {
    public ILookupParameter<DoubleMatrix> DistanceMatrixParameter {
      get {
        if (Parameters.ContainsKey("DistanceMatrix"))
          return (ILookupParameter<DoubleMatrix>)Parameters["DistanceMatrix"];
        else
          return null;
      }
    }

    [StorableConstructor]
    private VRPSimilarityCalculator(bool deserializing) : base(deserializing) { }
    private VRPSimilarityCalculator(VRPSimilarityCalculator original, Cloner cloner) : base(original, cloner) { }
    public VRPSimilarityCalculator() : base() {
      Parameters.Add(new LookupParameter<DoubleMatrix>("DistanceMatrix", "The matrix which contains the distances between the cities."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new VRPSimilarityCalculator(this, cloner);
    }

    public static double CalculateSimilarity(PotvinEncoding e1, PotvinEncoding e2) {
      // extract edges from first solution
      var edges1 = new List<Tuple<int, int>>();
      foreach (Tour tour in e1.Tours) {
        edges1.Add(new Tuple<int, int>(0, tour.Cities[0]));
        for (int i = 0; i < tour.Cities.Count - 1; i++)
          edges1.Add(new Tuple<int, int>(tour.Cities[i], tour.Cities[i + 1]));
        edges1.Add(new Tuple<int, int>(tour.Cities[tour.Cities.Count - 1], 0));
      }

      // extract edges from second solution
      var edges2 = new List<Tuple<int, int>>();
      foreach (Tour tour in e2.Tours) {
        edges2.Add(new Tuple<int, int>(0, tour.Cities[0]));
        for (int i = 0; i < tour.Cities.Count - 1; i++)
          edges2.Add(new Tuple<int, int>(tour.Cities[i], tour.Cities[i + 1]));
        edges2.Add(new Tuple<int, int>(tour.Cities[tour.Cities.Count - 1], 0));
      }

      if (edges1.Count + edges2.Count == 0)
        throw new ArgumentException("Cannot calculate diversity because no tours exist.");

      int identicalEdges = 0;
      foreach (var edge in edges1) {
        if (edges2.Any(x => x.Equals(edge)))
          identicalEdges++;
      }

      return identicalEdges * 2.0 / (edges1.Count + edges2.Count);
    }

    protected override double CalculateSimilarity(IScope left, IScope right) {
      IVRPEncoding e1 = left.Variables[TargetParameter.ActualName].Value as IVRPEncoding;
      if (!(e1 is PotvinEncoding))
        e1 = PotvinEncoding.ConvertFrom(e1 as IVRPEncoding, DistanceMatrixParameter);

      IVRPEncoding e2 = right.Variables[TargetParameter.ActualName].Value as IVRPEncoding;
      if (!(e2 is PotvinEncoding))
        e2 = PotvinEncoding.ConvertFrom(e2 as IVRPEncoding, DistanceMatrixParameter);

      return CalculateSimilarity(e1 as PotvinEncoding, e2 as PotvinEncoding);
    }
  }
}
